/* Copyright (C) 2008 Vincent Penquerc'h.
   This file is part of the Ogg/Kate codec library.
   Written by Vincent Penquerc'h.

   This file is preliminary code distributed only as a way to get
   feedback. As such, it is not meant to be finished quality, may
   not work, and streams encoded with it will certainly be unable
   to be decoded by any released version of the Ogg/Kate library.

   Since this code is not supposed to be used by third parties yet,
   there is no licence yet for this file, fair use applies.  */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <kate/kate.h>

int kate_comment_init(kate_comment *kc)
{
  if (!kc) return KE_INVALID_PARAMETER;
  kc->user_comments=NULL;
  kc->comment_lengths=NULL;
  kc->comments=0;
  kc->vendor=NULL;
  return 0;
}

int kate_comment_clear(kate_comment *kc)
{
  int n;

  if (!kc) return KE_INVALID_PARAMETER;

  for (n=0;n<kc->comments;++n) kate_free(kc->user_comments[n]);
  kate_free(kc->user_comments);
  kate_free(kc->comment_lengths);
  kate_free(kc->vendor);
  return 0;
}

int kate_comment_add(kate_comment *kc,const char *comment)
{
  int *cl;
  char **uc;
  size_t len;

  if (!kc || !comment) return KE_INVALID_PARAMETER;

  uc=kate_realloc(kc->user_comments,(kc->comments+1)*sizeof(char*));
  if (!uc) return KE_OUT_OF_MEMORY;
  kc->user_comments=uc;
  cl=kate_realloc(kc->comment_lengths,(kc->comments+1)*sizeof(int));
  if (!cl) return KE_OUT_OF_MEMORY;
  kc->comment_lengths=cl;

  len=strlen(comment);
  kc->user_comments[kc->comments]=(char*)kate_malloc(len+1);
  if (!kc->user_comments[kc->comments]) return KE_OUT_OF_MEMORY;
  strcpy(kc->user_comments[kc->comments],comment);
  kc->comment_lengths[kc->comments]=len;

  ++kc->comments;

  return 0;
}

int kate_comment_add_tag(kate_comment *kc,const char *name,const char *value)
{
  char *comment;

  if (!kc || !name || !value) return KE_INVALID_PARAMETER;

  comment=(char*)kate_malloc(strlen(name)+1+strlen(value)+1);
  sprintf("%s=%s",name,value);
  kate_comment_add(kc,comment);
  kate_free(comment);

  return 0;
}

const char *kate_comment_query(const kate_comment *kc,const char *tag,int count)
{
  int n;

  if (!kc) return NULL;

  for (n=0;n<kc->comments;++n) {
    char *eq=strchr(kc->user_comments[n],'=');
    if (!eq) continue; /* wrong format */
    if (!strncmp(tag,kc->user_comments[n],eq-kc->user_comments[n])) {
      if (count==0) return eq+1;
      --count;
    }
  }
  /* not found */
  return NULL;
}

int kate_comment_query_count(const kate_comment *kc,const char *tag)
{
  int n,count;

  if (!kc) return KE_INVALID_PARAMETER;

  count=0;
  for (n=0;n<kc->comments;++n) {
    char *eq=strchr(kc->user_comments[n],'=');
    if (!eq) continue; /* wrong format */
    if (!strncmp(tag,kc->user_comments[n],eq-kc->user_comments[n])) {
      ++count;
    }
  }

  return count;
}

