/* Copyright (C) 2008 Vincent Penquerc'h.
   This file is part of the Ogg/Kate codec library.
   Written by Vincent Penquerc'h.

   This file is preliminary code distributed only as a way to get
   feedback. As such, it is not meant to be finished quality, may
   not work, and streams encoded with it will certainly be unable
   to be decoded by any released version of the Ogg/Kate library.

   Since this code is not supposed to be used by third parties yet,
   there is no licence yet for this file, fair use applies.  */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ogg/ogg.h>
#include <kate/kate.h>
#include "kate_decode_state.h"

static void readbuf(oggpack_buffer *opb,char *s,int len)
{
  while (len--) *s++=oggpack_read(opb,8);
}

static void reads(oggpack_buffer *opb,char *s,int maxlen)
{
  while (maxlen--) {
    *s=oggpack_read(opb,8);
    if (!*s) break;
    ++s;
  }
}

static ogg_int32_t read32(oggpack_buffer *opb)
{
  ogg_int32_t v=0;
  v|=oggpack_read(opb,8);
  v|=(oggpack_read(opb,8)<<8);
  v|=(oggpack_read(opb,8)<<16);
  v|=(oggpack_read(opb,8)<<24);
  return v;
}

static double readf32(oggpack_buffer *opb)
{
  double v;
  char b[64];
  reads(opb,b,sizeof(b));
  b[sizeof(b)-1]=0;
  sscanf(b,"%lf",&v);
  return v;
}

static ogg_int64_t read64(oggpack_buffer *opb)
{
  ogg_int32_t vh=0,vl=0;
  vl|=oggpack_read(opb,8);
  vl|=(oggpack_read(opb,8)<<8);
  vl|=(oggpack_read(opb,8)<<16);
  vl|=(oggpack_read(opb,8)<<24);
  vh|=oggpack_read(opb,8);
  vh|=(oggpack_read(opb,8)<<8);
  vh|=(oggpack_read(opb,8)<<16);
  vh|=(oggpack_read(opb,8)<<24);
  return ((ogg_int64_t)vl)|(((ogg_int64_t)vh)<<32);
}

int kate_decode_is_header(const ogg_packet *op)
{
  oggpack_buffer opb;
  unsigned char headerid;
  char magic[8];

  if (!op) return 0;

  if (op->packetno!=0) return 0;

  oggpack_readinit(&opb,op->packet,op->bytes);
  headerid=oggpack_read(&opb,8);
  if (headerid!=0x80) return 0;
  readbuf(&opb,magic,8);
  if (memcmp(magic,"kate\0\0\0\0",8)) return 0;
  
  return 1;
}

static int kate_decode_info_header(kate_info *ki,oggpack_buffer *opb)
{
  char magic[8];
  int len;
  
  readbuf(opb,magic,8);
  if (memcmp(magic,"kate\0\0\0\0",8)) return KE_BAD_PACKET;

  ki->bitstream_version_major=oggpack_read(opb,8);
  ki->bitstream_version_minor=oggpack_read(opb,8);
  ki->gps_numerator=read32(opb);
  ki->gps_denominator=read32(opb);
  ki->text_encoding=oggpack_read(opb,8);

  if (ki->gps_denominator==0) return KE_BAD_PACKET;

  len=read32(opb);
  ki->language=(char*)kate_malloc(len+1);
  if (!ki->language) return KE_OUT_OF_MEMORY;
  if (len) {
    readbuf(opb,ki->language,len+1); /* a terminating null is included */
    if (ki->language[len]) return KE_BAD_PACKET; /* but do check it to be sure */
  }
  ki->language[len]=0;

  len=read32(opb);
  ki->category=(char*)kate_malloc(len+1);
  if (!ki->category) return KE_OUT_OF_MEMORY;
  if (len) {
    readbuf(opb,ki->category,len+1); /* a terminating null is included */
    if (ki->category[len]) return KE_BAD_PACKET; /* but do check it to be sure */
  }
  ki->category[len]=0;

  if (oggpack_read(opb,1)!=1) return KE_BAD_PACKET; /* EOP check */

  return 0;
}

static int kate_decode_comment_packet(kate_comment *kc,oggpack_buffer *opb)
{
  int len,nc;
  char *s;

  if (!kc || !opb) return KE_INVALID_PARAMETER;

  len=read32(opb);
  if (len<0) return KE_BAD_PACKET;
  s=(char*)kate_malloc(len+1);
  if (!s) return KE_OUT_OF_MEMORY;
  kc->vendor=s;
  readbuf(opb,kc->vendor,len);
  kc->vendor[len]=0;

  nc=read32(opb);
  if (nc<0) return KE_BAD_PACKET;
  kc->user_comments=(char**)kate_malloc(nc*sizeof(char*));
  kc->comment_lengths=(int*)kate_malloc(nc*sizeof(int));
  if (!kc->user_comments || !kc->comment_lengths) return KE_OUT_OF_MEMORY;

  kc->comments=nc;
  for (nc=0;nc<kc->comments;++nc) kc->user_comments[nc]=NULL;
  for (nc=0;nc<kc->comments;++nc) {
    len=read32(opb);
    if (len<0) return KE_BAD_PACKET;
    kc->user_comments[nc]=(char*)kate_malloc(len+1);
    if (!kc->user_comments[nc]) return KE_BAD_PACKET;
    if (len) {
      readbuf(opb,kc->user_comments[nc],len);
    }
    kc->user_comments[nc][len]=0;
    kc->comment_lengths[nc]=len;
  }

  if (oggpack_read(opb,1)!=1) return KE_BAD_PACKET; /* EOP check */

  return 0;
}

static int kate_decode_regions_packet(kate_info *ki,oggpack_buffer *opb)
{
  int n,nregions;
  kate_region *regions,*kr;

  if (!ki || !opb) return KE_INVALID_PARAMETER;

  nregions=read32(opb);
  if (nregions<0) return KE_BAD_PACKET;
  regions=(kate_region*)kate_malloc(nregions*sizeof(kate_region));
  if (!regions) return KE_OUT_OF_MEMORY;
  kr=regions;
  for (n=0;n<nregions;++n,++kr) {
    kr->metric=oggpack_read(opb,8);
    kr->x=read32(opb);
    kr->y=read32(opb);
    kr->w=read32(opb);
    kr->h=read32(opb);
    kr->style=read32(opb);
  }

  if (oggpack_read(opb,1)!=1) {
    kate_free(regions);
    return KE_BAD_PACKET; /* EOP check */
  }

  ki->nregions=nregions;
  ki->regions=regions;

  return 0;
}

static int kate_decode_color(kate_color *kc,oggpack_buffer *opb)
{
  if (!kc || !opb) return KE_INVALID_PARAMETER;
  kc->r=oggpack_read(opb,8);
  kc->g=oggpack_read(opb,8);
  kc->b=oggpack_read(opb,8);
  kc->a=oggpack_read(opb,8);
  return 0;
}

static int kate_decode_styles_packet(kate_info *ki,oggpack_buffer *opb)
{
  int n,nstyles;
  kate_style *styles,*ks;

  if (!ki || !opb) return KE_INVALID_PARAMETER;

  nstyles=read32(opb);
  if (nstyles<0) return KE_BAD_PACKET;
  styles=(kate_style*)kate_malloc(nstyles*sizeof(kate_style));
  if (!styles) return KE_OUT_OF_MEMORY;
  ks=styles;
  for (n=0;n<nstyles;++n,++ks) {
    ks->halign=oggpack_read(opb,2);
    ks->valign=oggpack_read(opb,2);
    kate_decode_color(&ks->text_color,opb);
    kate_decode_color(&ks->background_color,opb);
  }

  if (oggpack_read(opb,1)!=1) {
    kate_free(styles);
    return KE_BAD_PACKET; /* EOP check */
  }

  ki->nstyles=nstyles;
  ki->styles=styles;

  return 0;
}

static int kate_decode_curve(kate_curve *kc,oggpack_buffer *opb)
{
  size_t n;

  if (!kc || !opb) return KE_INVALID_PARAMETER;

  kc->type=oggpack_read(opb,8);
  kc->npts=read32(opb);
  kc->pts=(double*)kate_malloc(kc->npts*2*sizeof(double));
  if (!kc->pts) return KE_OUT_OF_MEMORY;
  for (n=0;n<kc->npts*2;++n) {
    kc->pts[n]=readf32(opb);
  }

  return 0;
}

static int kate_decode_motion(kate_motion *km,oggpack_buffer *opb)
{
  size_t n;

  if (!km || !opb) return KE_INVALID_PARAMETER;

  km->ncurves=read32(opb);
  km->curves=(kate_curve*)kate_malloc(km->ncurves*sizeof(kate_curve));
  km->durations=(double*)kate_malloc(km->ncurves*sizeof(double));
  if (!km->durations || !km->curves) return KE_OUT_OF_MEMORY;
  for (n=0;n<km->ncurves;++n) {
    int ret=kate_decode_curve(&km->curves[n],opb);
    if (ret<0) {
      kate_free(km->curves);
      kate_free(km->durations);
      return ret;
    }
    km->durations[n]=readf32(opb);
  }
  km->mapping=oggpack_read(opb,8);

  return 0;
}

int kate_decode_headerin(kate_info *ki,kate_comment *kc,ogg_packet *op)
{
  oggpack_buffer opb;
  unsigned char headerid;
  int ret=0;
  int done=0;

  if (!ki || !kc || !op) return KE_INVALID_PARAMETER;

  oggpack_readinit(&opb,op->packet,op->bytes);
  headerid=oggpack_read(&opb,8);

  switch (op->packetno) {
    case 0: /* this is the info packet */
      if (headerid!=0x80) return KE_BAD_PACKET;
      if (ki->gps_denominator) return KE_BAD_PACKET; /* already seen */
      ret=kate_decode_info_header(ki,&opb);
      break;

    case 1: /* this is the comments packet */
      if (headerid!=0x81) return KE_BAD_PACKET;
      if (kc->vendor) return KE_BAD_PACKET; /* already seen */
      ret=kate_decode_comment_packet(kc,&opb);
      break;

    case 2: /* this is the region list packet */
      if (headerid!=0x82) return KE_BAD_PACKET;
      if (ki->nregions!=0) return KE_BAD_PACKET; /* already seen */
      ret=kate_decode_regions_packet(ki,&opb);
      break;

    case 3: /* this is the style list packet */
      if (headerid!=0x83) return KE_BAD_PACKET;
      if (ki->nstyles!=0) return KE_BAD_PACKET; /* already seen */
      ret=kate_decode_styles_packet(ki,&opb);
      done=1; /* we're done, we know of no more headers to come */
      break;

    default:
      /* we ignore extra header packets for future proofing */
      if (!(headerid&0x80)) return KE_BAD_PACKET;
      ret=0;
      break;
  }

  if (!ret && done) {
    /* if we're correctly parsed the last packet, return positive
       instead of zero */
    ret=1;
  }

  return ret;
}

int kate_decode_init(kate_state *k,kate_info *ki)
{
  if (!k || !ki) return KE_INVALID_PARAMETER;

  k->ki=ki;
  k->granulepos=-1;
  k->packetno=-1;
  k->kds=kate_decode_state_create(ki);
  k->kes=NULL;

  return 0;
}

#define READ_OVERRIDE(read) \
  do { \
    if (oggpack_read(opb,1)) { read; } \
  } while(0)

static int kate_decode_text_packet(kate_state *k,oggpack_buffer *opb)
{
  int ret;
  int len;
  char *text;
  kate_decode_state *kds;
  kate_event *ev;

  if (!k || !opb) return KE_INVALID_PARAMETER;
  if (!k->kds) return KE_INIT;

  kds=k->kds;
  ret=kate_decode_state_clear(kds,k->ki,1);
  if (ret<0) return ret;

  ev=&kds->event;

  if (oggpack_read(opb,1)) {
    kate_decode_motion(&ev->motion,opb);
  }

  if (oggpack_read(opb,1)) {
    READ_OVERRIDE(ev->text_encoding=oggpack_read(opb,8));
    READ_OVERRIDE(
      do {
        len=read32(opb);
        if (len<0) return KE_BAD_PACKET;
        if (len>0) {
          ev->language=(char*)kate_malloc(len+1);
          readbuf(opb,ev->language,len);
          ev->language[len]=0;
        }
      } while(0)
    );
    READ_OVERRIDE(ev->region=read32(opb));
    READ_OVERRIDE(ev->style=read32(opb));
  }

  ev->end_granulepos=read64(opb);

  len=read32(opb);
  if (len<0) return KE_BAD_PACKET;

  /* read the text, and null terminate it - 4 characters for UTF-32 */
  text=(char*)kate_malloc(len+4);
  if (!text) return KE_BAD_PACKET;
  readbuf(opb,text,len);
  text[len]=0;
  text[len+1]=0;
  text[len+2]=0;
  text[len+3]=0;

  ev->text=text;
  ev->len=len;

  return 0;
}

static int kate_decode_end_packet(kate_state *k,oggpack_buffer *opb)
{
  int ret;
  kate_decode_state *kds;

  if (!k || !opb) return KE_INVALID_PARAMETER;
  if (!k->kds) return KE_INIT;

  kds=k->kds;
  ret=kate_decode_state_clear(kds,k->ki,1);
  if (ret<0) return ret;

  return 1;
}

int kate_decode_packetin(kate_state *k,ogg_packet *op)
{
  oggpack_buffer opb;
  int id;

  if (!k || !op) return KE_INVALID_PARAMETER;
  if (!k->ki) return KE_INIT;
  if (!k->kds) return KE_INIT;

  oggpack_readinit(&opb,op->packet,op->bytes);
  id=oggpack_read(&opb,8);
  if (id&0x80) {
    /* we have a header - we'll ignore it
       it may have happened either because we seeked to the beginning of the stream
       or because we're parsing a stream that has more headers than we know about */
    return 0;
  }

  switch (id) {
    case 0x00: return kate_decode_text_packet(k,&opb);
    case 0x01: return kate_decode_end_packet(k,&opb);
    default: return 0; /* unknown data packets are ignored, for future proofing */
  }

  return 0;
}

int kate_decode_eventout(kate_state *k,const kate_event **event)
{
  if (!k) return KE_INVALID_PARAMETER;
  if (!k->kds) return KE_INIT;

  if (event) *event=&k->kds->event;

  return 0;
}

