/* Copyright (C) 2008 Vincent Penquerc'h.
   This file is part of the Ogg/Kate codec library.
   Written by Vincent Penquerc'h.

   This file is preliminary code distributed only as a way to get
   feedback. As such, it is not meant to be finished quality, may
   not work, and streams encoded with it will certainly be unable
   to be decoded by any released version of the Ogg/Kate library.

   Since this code is not supposed to be used by third parties yet,
   there is no licence yet for this file, fair use applies.  */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <kate/kate.h>
#include "kate_granule.h"
#include "kate_encode_state.h"

int kate_encode_init(kate_state *k,kate_info *ki)
{
  if (!k || !ki) return KE_INVALID_PARAMETER;

  k->ki=ki;
  k->granulepos=0;
  k->packetno=0;
  k->kes=kate_encode_state_create();
  k->kds=NULL;

  return 0;
}

static void writebuf(oggpack_buffer *opb,const char *s,int len)
{
  while (len--) oggpack_write(opb,*s++,8);
}

static void writes(oggpack_buffer *opb,const char *s)
{
  do oggpack_write(opb,*s,8); while (*s++);
}

static void write32(oggpack_buffer *opb,ogg_int32_t v)
{
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
}

static void writef32(oggpack_buffer *opb,double v)
{
  char b[64];
  snprintf(b,sizeof(b),"%f",v);
  b[sizeof(b)-1]=0;
  writes(opb,b);
}

static void write64(oggpack_buffer *opb,ogg_int64_t v)
{
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
  v>>=8;
  oggpack_write(opb,v&0xff,8);
}

static int kate_finalize_packet_buffer(oggpack_buffer *opb,ogg_packet *op,kate_state *k,int last)
{
  if (!opb || !op || !k) return KE_INVALID_PARAMETER;

  /* a little bit to mark the end, so decode can check it */
  oggpack_write(opb,1,1);

  op->bytes=oggpack_bytes(opb);
  /* use _ogg_malloc here as it will be owned by Ogg */
  op->packet=_ogg_malloc(op->bytes);
  if (!op->packet) return KE_OUT_OF_MEMORY;

  memcpy(op->packet,oggpack_get_buffer(opb),op->bytes);

  /* reset the buffer so we're ready for next packet */
  oggpack_writeclear(opb);
  oggpack_writeinit(opb);

  op->b_o_s=0;
  op->e_o_s=last;
  op->packetno=k->packetno;
  op->granulepos=k->granulepos;

  ++k->packetno;

  /* clear any overrides */
  kate_encode_state_clear_overrides(k->kes,k->ki);

  return 0;
}

static int kate_encode_info(kate_state *k,ogg_packet *op)
{
  oggpack_buffer *opb;
  kate_info *ki;

  if (!k || !op) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;

  opb=&k->kes->opb;
  oggpack_write(opb,0x80,8);

  writebuf(opb,"kate\0\0\0\0",8);

  ki=k->ki;
  oggpack_write(opb,KATE_BITSTREAM_VERSION_MAJOR,8);
  oggpack_write(opb,KATE_BITSTREAM_VERSION_MINOR,8);
  write32(opb,ki->gps_numerator);
  write32(opb,ki->gps_denominator);
  oggpack_write(opb,ki->text_encoding,8);

  if (ki->language) {
    size_t len=strlen(ki->language);
    write32(opb,len);
    writebuf(opb,ki->language,len+1); /* the terminating null is written */
  }
  else write32(opb,0);

  if (ki->category) {
    size_t len=strlen(ki->category);
    write32(opb,len);
    writebuf(opb,ki->category,len+1); /* the terminating null is written */
  }
  else write32(opb,0);

  return kate_finalize_packet_buffer(opb,op,k,0);
}

static int kate_encode_comment(kate_state *k,kate_comment *kc,ogg_packet *op)
{
  oggpack_buffer *opb;

  if (!k || !kc || !op) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;

  opb=&k->kes->opb;
  oggpack_write(opb,0x81,8);

  const char *vendor=kate_get_version_string();
  int vendor_len=strlen(vendor);

  /* mostly copied from theora encoder_toplevel.c */
  write32(opb,vendor_len);
  writebuf(opb,vendor,vendor_len);

  write32(opb,kc->comments);
  if (kc->comments) {
    int i;
    for(i=0;i<kc->comments;++i) {
      if (kc->user_comments[i]) {
        write32(opb,kc->comment_lengths[i]);
        writebuf(opb,kc->user_comments[i],kc->comment_lengths[i]);
      }
      else {
        oggpack_write(opb,0,32);
      }
    }
  }
  
  return kate_finalize_packet_buffer(opb,op,k,0);
}

static int kate_encode_region(const kate_region *kr,oggpack_buffer *opb)
{
  if (!kr || !opb) return KE_INVALID_PARAMETER;

  oggpack_write(opb,kr->metric,8);
  write32(opb,kr->x);
  write32(opb,kr->y);
  write32(opb,kr->w);
  write32(opb,kr->h);
  write32(opb,kr->style);

  return 0;
}

static int kate_encode_regions(kate_state *k,ogg_packet *op)
{
  oggpack_buffer *opb;
  kate_info *ki;
  size_t i;
  int ret;

  if (!k || !op) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;

  opb=&k->kes->opb;
  oggpack_write(opb,0x82,8);

  ki=k->ki;
  if (!ki) return KE_INIT;

  write32(opb,ki->nregions>0?ki->nregions:1);

  /* if no regions, place a default one covering the entire area */
  if (ki->nregions==0) {
    kate_region full={kate_percentage,0,0,100,100,0};
    ret=kate_encode_region(&full,opb);
    if (ret<0) return ret;
  }
  else {
    const kate_region *kr=ki->regions;
    for(i=0;i<ki->nregions;++i,++kr) {
      ret=kate_encode_region(kr,opb);
      if (ret<0) return ret;
    }
  }
  
  return kate_finalize_packet_buffer(opb,op,k,0);
}

static int kate_encode_color(const kate_color *kc,oggpack_buffer *opb)
{
  if (!kc || !opb) return KE_INVALID_PARAMETER;
  oggpack_write(opb,kc->r,8);
  oggpack_write(opb,kc->g,8);
  oggpack_write(opb,kc->b,8);
  oggpack_write(opb,kc->a,8);
  return 0;
}

static int kate_encode_style(const kate_style *ks,oggpack_buffer *opb)
{
  if (!ks || !opb) return KE_INVALID_PARAMETER;

  oggpack_write(opb,ks->halign,2);
  oggpack_write(opb,ks->valign,2);
  kate_encode_color(&ks->text_color,opb);
  kate_encode_color(&ks->background_color,opb);

  return 0;
}

static int kate_encode_styles(kate_state *k,ogg_packet *op)
{
  oggpack_buffer *opb;
  kate_info *ki;
  size_t i;
  int ret;

  if (!k || !op) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;

  opb=&k->kes->opb;
  oggpack_write(opb,0x83,8);

  ki=k->ki;
  if (!ki) return KE_INIT;

  write32(opb,ki->nstyles>0?ki->nstyles:1);

  /* if no styles, place a default one with sensible defaults */
  if (ki->nstyles==0) {
    kate_style full={kate_hcenter,kate_vcenter,{255,255,255,255},{0,0,0,0}};
    ret=kate_encode_style(&full,opb);
    if (ret<0) return ret;
  }
  else {
    const kate_style *ks=ki->styles;
    for(i=0;i<ki->nstyles;++i,++ks) {
      ret=kate_encode_style(ks,opb);
      if (ret<0) return ret;
    }
  }
  
  return kate_finalize_packet_buffer(opb,op,k,0);
}

static int kate_encode_curve(const kate_curve *kc,oggpack_buffer *opb)
{
  size_t n;

  if (!kc || !opb) return KE_INVALID_PARAMETER;
  
  oggpack_write(opb,kc->type,8);
  write32(opb,kc->npts);
  for (n=0;n<kc->npts*2;++n) {
    writef32(opb,kc->pts[n]);
  }

  return 0;
}

static int kate_encode_motion(const kate_motion *km,oggpack_buffer *opb)
{
  size_t n;

  if (!km || !opb) return KE_INVALID_PARAMETER;
  
  write32(opb,km->ncurves);
  for (n=0;n<km->ncurves;++n) {
    int ret=kate_encode_curve(&km->curves[n],opb);
    if (ret<0) return ret;
    writef32(opb,km->durations[n]);
  }
  oggpack_write(opb,km->mapping,8);

  return 0;
}

static int kate_twiddle_granule(kate_state *k,ogg_int64_t *granulepos)
{
  if (granule2granule(k->ki,*granulepos)<granule2granule(k->ki,k->granulepos)) return -1;
  if (*granulepos<=k->granulepos) *granulepos=k->granulepos+1;
  return 0;
}

#define WRITE_OVERRIDE(var,def,write) \
  do \
    if (kes->overrides.var!=(def)) { \
      oggpack_write(opb,1,1); \
      write; \
    } \
    else oggpack_write(opb,0,1); \
  while(0)

static int kate_encode_overrides(kate_state *k,oggpack_buffer *opb)
{
  kate_encode_state *kes;
  if (!k || !opb) return KE_INVALID_PARAMETER;
  kes=k->kes;
  if (!kes) return KE_INIT;

  if (kes->overrides.language
   || kes->overrides.text_encoding!=k->ki->text_encoding
   || kes->overrides.region
   || kes->overrides.style
  ) {
    oggpack_write(opb,1,1);
    WRITE_OVERRIDE(text_encoding,k->ki->text_encoding,oggpack_write(opb,kes->overrides.text_encoding,8));
    WRITE_OVERRIDE(language,NULL,
        do {
      size_t len=strlen(kes->overrides.language);
      write32(opb,len);
      writebuf(opb,kes->overrides.language,len);
      } while(0));
    WRITE_OVERRIDE(region,-1,write32(opb,kes->overrides.region));
    WRITE_OVERRIDE(style,-1,write32(opb,kes->overrides.style));
  }
  else oggpack_write(opb,0,1);

  return 0;
}

int kate_encode_text(kate_state *k,double t0,double t1,const kate_motion *km,const char *text,size_t sz,ogg_packet *op)
{
  oggpack_buffer *opb;
  ogg_int64_t granulepos,end_granulepos;
  int ret;

  if (!k || !text || !op) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;

  granulepos=kate_time_granule(k->ki,t0);
  if (kate_twiddle_granule(k,&granulepos)<0) return KE_BAD_GRANULE;

  end_granulepos=kate_time_granule(k->ki,t1);
  
  opb=&k->kes->opb;
  oggpack_write(opb,0x00,8);

  if (km) {
    oggpack_write(opb,1,1);
    ret=kate_encode_motion(km,opb);
    if (ret<0) return ret;
  }
  else oggpack_write(opb,0,1);

  kate_encode_overrides(k,opb);

  write64(opb,end_granulepos);

  write32(opb,sz);
  writebuf(opb,text,sz);

  if (granulepos>k->kes->furthest_granule) k->kes->furthest_granule=granulepos;
  if (end_granulepos>k->kes->furthest_granule) k->kes->furthest_granule=end_granulepos;

  k->granulepos=granulepos;
  return kate_finalize_packet_buffer(opb,op,k,0);
}

int kate_encode_finish(kate_state *k,ogg_packet *op)
{
  oggpack_buffer *opb;
  ogg_int64_t end_granulepos;

  if (!k || !op) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;

  end_granulepos=k->kes->furthest_granule;
  if (kate_twiddle_granule(k,&end_granulepos)<0) return KE_BAD_GRANULE;
  k->granulepos=end_granulepos;

  opb=&k->kes->opb;
  oggpack_write(opb,0x01,8);

  return kate_finalize_packet_buffer(opb,op,k,1);
}

int kate_encode_headers(kate_state *k,kate_comment *kc,ogg_packet *op)
{
  if (!k || !kc || !op) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;

  switch (k->packetno) {
    case 0: return kate_encode_info(k,op);
    case 1: return kate_encode_comment(k,kc,op);
    case 2: return kate_encode_regions(k,op);
    case 3: return kate_encode_styles(k,op);
    case 4: return 1;
    default: return KE_INVALID_PARAMETER;
  }
}

int kate_encode_set_region(kate_state *k,size_t region)
{
  if (!k) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;
  if (!k->ki) return KE_INIT;
  if (region>=k->ki->nregions) return KE_INVALID_PARAMETER;
  k->kes->overrides.region=region;
  return 0;
}

int kate_encode_set_style(kate_state *k,size_t style)
{
  if (!k) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;
  if (!k->ki) return KE_INIT;
  if (style>=k->ki->nstyles) return KE_INVALID_PARAMETER;
  k->kes->overrides.style=style;
  return 0;
}

int kate_encode_set_text_encoding(kate_state *k,kate_text_encoding text_encoding)
{
  if (!k) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;
  k->kes->overrides.text_encoding=text_encoding;
  return 0;
}

int kate_encode_set_language(kate_state *k,const char *language) /* language can be NULL */
{
  size_t len;
  char *l=NULL;

  if (!k) return KE_INVALID_PARAMETER;
  if (!k->kes) return KE_INIT;

  if (language) {
    len=strlen(language);
    l=(char*)kate_malloc(len+1);
    if (!l) return KE_OUT_OF_MEMORY;
    memcpy(l,language,len+1);
  }

  if (k->kes->overrides.language) kate_free(k->kes->overrides.language);
  k->kes->overrides.language=l;

  return 0;
}

