/* Copyright (C) 2008 Vincent Penquerc'h.
   This file is part of the Ogg/Kate codec library.
   Written by Vincent Penquerc'h.

   This file is preliminary code distributed only as a way to get
   feedback. As such, it is not meant to be finished quality, may
   not work, and streams encoded with it will certainly be unable
   to be decoded by any released version of the Ogg/Kate library.

   Since this code is not supposed to be used by third parties yet,
   there is no licence yet for this file, fair use applies.  */


#include <stdlib.h>
#include <string.h>
#include <kate/kate.h>

int kate_info_init(kate_info *ki)
{
  if (!ki) return KE_INVALID_PARAMETER;

  ki->bitstream_version_major=KATE_BITSTREAM_VERSION_MAJOR;
  ki->bitstream_version_minor=KATE_BITSTREAM_VERSION_MINOR;
  ki->gps_numerator=0;
  ki->gps_denominator=0;
  ki->text_encoding=kate_utf8;
  ki->language=NULL;
  ki->category=NULL;
  ki->nregions=0;
  ki->regions=NULL;
  ki->nstyles=0;
  ki->styles=NULL;

  return 0;
}

static int kate_replace_string(char **sptr,const char *s)
{
  size_t len;
  char *l=NULL;

  if (!sptr) return KE_INVALID_PARAMETER;

  if (s) {
    len=strlen(s);
    l=(char*)kate_malloc(len+1);
    if (!l) return KE_OUT_OF_MEMORY;
    memcpy(l,s,len+1);
  }

  if (*sptr) kate_free(*sptr);
  *sptr=l;

  return 0;
}

int kate_info_set_language(kate_info *ki,const char *language)
{
  if (!ki) return KE_INVALID_PARAMETER;
  return kate_replace_string(&ki->language,language);
}

int kate_info_set_category(kate_info *ki,const char *category)
{
  if (!ki) return KE_INVALID_PARAMETER;
  return kate_replace_string(&ki->category,category);
}

int kate_info_add_region(kate_info *ki,const kate_region *kr)
{
  if (!ki || !kr) return KE_INVALID_PARAMETER;

  kate_region *regions=(kate_region*)kate_realloc(ki->regions,(ki->nregions+1)*sizeof(kate_region));
  if (!regions) return KE_OUT_OF_MEMORY;
  ki->regions=regions;
  ki->regions[ki->nregions++]=*kr;

  return 0;
}

int kate_info_add_style(kate_info *ki,const kate_style *ks)
{
  if (!ki || !ks) return KE_INVALID_PARAMETER;

  kate_style *styles=(kate_style*)kate_realloc(ki->styles,(ki->nstyles+1)*sizeof(kate_style));
  if (!styles) return KE_OUT_OF_MEMORY;
  ki->styles=styles;
  ki->styles[ki->nstyles++]=*ks;

  return 0;
}

int kate_info_matches_language(const kate_info *ki,const char *language)
{
  char *sep;
  if (!ki) return KE_INVALID_PARAMETER;
  
  if (!language) return 1; /* if we ask for "any" language, we match */
  if (!ki->language) return 1; /* if the stream has no set language, it matches all */

  if (!strcasecmp(ki->language,language)) return 1; /* perfect match */

  /* if we specify a language with no subtag, it matches any subtag of the same language */
  sep=strpbrk(language,"-_");
  if (sep) {
    if (!strncasecmp(ki->language,language,sep-language)) return 2; /* partial match */
  }

  return 0; /* doesn't match */
}

int kate_info_clear(kate_info *ki)
{
  if (!ki) return KE_INVALID_PARAMETER;

  if (ki->regions) kate_free(ki->regions);
  if (ki->styles) kate_free(ki->styles);
  if (ki->language) kate_free(ki->language);
  if (ki->category) kate_free(ki->category);

  return 0;
}

