/* Copyright (C) 2008 Vincent Penquerc'h.
   This file is part of the Ogg/Kate codec library.
   Written by Vincent Penquerc'h.

   This file is preliminary code distributed only as a way to get
   feedback. As such, it is not meant to be finished quality, may
   not work, and streams encoded with it will certainly be unable
   to be decoded by any released version of the Ogg/Kate library.

   Since this code is not supposed to be used by third parties yet,
   there is no licence yet for this file, fair use applies.  */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ogg/ogg.h>
#include <kate/kate.h>

static ogg_sync_state sync;
static ogg_stream_state os;
static ogg_page og;
static ogg_packet op;
static kate_state k;
static kate_info ki;
static kate_comment kc;

enum { uninitialized, header_info, header_comment, data };

int main()
{
  size_t read;
  int ret;
  int eos=0;
  int init=uninitialized;

  ogg_sync_init(&sync);
  
  while (1) {
    char *buffer=ogg_sync_buffer(&sync,4096);
    if (!buffer) {
      fprintf(stderr,"Failed to get sync buffer\n");
      break;
    }
    read=fread(buffer,1,4096,stdin);
    if (read==0) eos=1;
    ogg_sync_wrote(&sync,read);

    while (ogg_sync_pageout(&sync,&og)) {
      if (ogg_page_bos(&og)) {
        if (init==uninitialized) {
          ogg_stream_init(&os,ogg_page_serialno(&og));
          kate_info_init(&ki);
          kate_comment_init(&kc);
          init=header_info;
        }
      }
      ret=ogg_stream_pagein(&os,&og);
      if (ret>=0) {
        while (ogg_stream_packetout(&os,&op)) {
          if (init<data) {
            int ret=kate_decode_headerin(&ki,&kc,&op);
            if (ret>=0) {
              /* we found a Kate bitstream */
              if (op.packetno==0) {
                printf("Bitstream %08x looks like Kate\n",ogg_page_serialno(&og));
              }
              if (ret>0) {
                /* we're done parsing headers, go for data */
                printf("Bitstream %08x is Kate (\"%s\", encoding %d)\n",ogg_page_serialno(&og),ki.language,ki.text_encoding);
                kate_decode_init(&k,&ki);
                init=data;
              }
            }
            else {
              if (init!=uninitialized) {
                kate_info_clear(&ki);
                kate_comment_clear(&kc);
                init=uninitialized;
              }
            }
          }
          else {
            int ret=kate_decode_packetin(&k,&op);
            if (ret<0) {
              printf("error in kate_decode_packetin: %d\n",ret);
            }
            else if (ret>0) {
              /* we're done */
            }
            else {
              const kate_event *ev=NULL;
              ret=kate_decode_eventout(&k,&ev);
              if (ret<0) {
                printf("error in kate_decode_eventout: %d\n",ret);
              }
              else {
                ogg_int64_t granule=ogg_page_granulepos(&og);
                printf("Got text at %f [r%d s%d] (%d bytes): %s\n",kate_granule_time(&ki,granule),ev->region,ev->style,ev->len,ev->text);
                if (ev->motion.ncurves) {
                  size_t s;
                  double t,duration;
                  printf(" + motion, %u curves (",ev->motion.ncurves);
                  for (s=0;s<ev->motion.ncurves;++s) printf("%s%f",s?", ":"",ev->motion.durations[s]);
                  printf(")\n");

                  duration=0.0;
                  for (s=0;s<ev->motion.ncurves;++s) duration+=ev->motion.durations[s];
                  for (t=0;t<duration;t+=duration/100.0) {
                    double x,y;
                    kate_get_motion_point(&ev->motion,t,&x,&y);
                    printf("%f: %f %f\n",t,x,y);
                  }
                }
              }
            }
          }
        }
      }
    }

    if (eos) break;
  }

  ogg_stream_clear(&os);
  ogg_sync_clear(&sync);

  return 0;
}
