/* Copyright (C) 2008 Vincent Penquerc'h.
   This file is part of the Ogg/Kate codec library.
   Written by Vincent Penquerc'h.

   This file is preliminary code distributed only as a way to get
   feedback. As such, it is not meant to be finished quality, may
   not work, and streams encoded with it will certainly be unable
   to be decoded by any released version of the Ogg/Kate library.

   Since this code is not supposed to be used by third parties yet,
   there is no licence yet for this file, fair use applies.  */


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ogg/ogg.h>
#include <kate/kate.h>

static ogg_stream_state os;
static ogg_page og;
static ogg_packet op;
static kate_state k;
static kate_info ki;
static kate_comment kc;
static char str[4096];
static int headers_written=0;

static double test_control_points[]={0,0, 0,1, 1,1, 1,1};
static kate_curve test_curves[]={
  {kate_curve_linear_interpolation,4,test_control_points},
  {kate_curve_linear_interpolation,4,test_control_points},
  {kate_curve_linear_interpolation,4,test_control_points},
  {kate_curve_linear_interpolation,4,test_control_points},
};
static double test_curve_durations[]={1.0, 0.5, 0.25, 0.75};

static kate_motion test_motion={
  4,
  test_curves,
  test_curve_durations,
  kate_motion_mapping_text
};
static int use_motion=0;

static void flush_page()
{
  while (1) {
    int ret=ogg_stream_flush(&os,&og);
    if (ret==0) break;
    fwrite(og.header,1,og.header_len,stdout);
    fwrite(og.body,1,og.body_len,stdout);
  }
}

static void poll_page()
{
  /*
  int ret=ogg_stream_pageout(&os,&og);
  if (ret==0) return;
  fwrite(og.header,1,og.header_len,stdout);
  fwrite(og.body,1,og.body_len,stdout);
  */

  /* discontinuous codec, so we want to flush pages every time,
     as Ogg keeps granules per page, not per packet */
  flush_page();
}

static void write_headers()
{
  kate_comment_init(&kc);
  kate_comment_add(&kc,"test_title=test_value");

  while (1) {
    int ret=kate_encode_headers(&k,&kc,&op);
    if (ret<0) {
      fprintf(stderr,"error encoding headers: %d\n",ret);
      break;
    }
    if (ret>0) break; /* we're done */
    ogg_stream_packetin(&os,&op);

    /* place all headers on the same page, it keeps oggmerge happy */
    /* poll_page(); */
  }

  flush_page();

  kate_comment_clear(&kc);
}

static int starts(const char *s,const char *start)
{
  return !strncmp(s,start,strlen(start));
}

static void flush_headers()
{
  if (!headers_written) {
    write_headers();
    headers_written=1;
  }
}

int main()
{
  kate_info_init(&ki);
  ki.gps_numerator=1000;
  ki.gps_denominator=1;

  kate_encode_init(&k,&ki);

  srand(time(NULL)^getpid());
  ogg_stream_init(&os,rand());

  fgets(str,sizeof(str),stdin);
  while (!feof(stdin)) {
    if (str[0]==':') {
      if (starts(str,":language:")) {
        char language[64];
        sscanf(str,":language: %s\n",language);
        kate_info_set_language(&ki,language);
      }
      else if (starts(str,":def style:")) {
        kate_style ks;
        static char halign[256],valign[256],r0[256],g0[256],b0[256],a0[256],r1[256],g1[256],b1[256],a1[256];
        sscanf(str,":def style: %s %s text %s %s %s %s bg %s %s %s %s\n",halign,valign,r0,g0,b0,a0,r1,g1,b1,a1);
        if (!strcmp(halign,"hleft")) ks.halign=kate_hleft;
        else if (!strcmp(halign,"hcenter")) ks.halign=kate_hcenter;
        else if (!strcmp(halign,"hright")) ks.halign=kate_hright;
        else if (!strcmp(halign,"hwhole")) ks.halign=kate_hwhole;
        else fprintf(stderr,"Invalid halign: %s\n",halign);
        if (!strcmp(valign,"vtop")) ks.halign=kate_vtop;
        else if (!strcmp(valign,"vcenter")) ks.halign=kate_vcenter;
        else if (!strcmp(valign,"vbottom")) ks.halign=kate_vbottom;
        else if (!strcmp(valign,"vwhole")) ks.halign=kate_vwhole;
        else fprintf(stderr,"Invalid valign: %s\n",valign);
        ks.text_color.r=atoi(r0);
        ks.text_color.g=atoi(g0);
        ks.text_color.b=atoi(b0);
        ks.text_color.a=atoi(a0);
        ks.background_color.r=atoi(r1);
        ks.background_color.g=atoi(g1);
        ks.background_color.b=atoi(b1);
        ks.background_color.a=atoi(a1);
        kate_info_add_style(&ki,&ks);
      }
      else if (starts(str,":def region:")) {
        kate_region kr;
        static char metric[256],x[256],y[256],w[256],h[256],style[256];
        sscanf(str,":def region: %s %s %s %s %s style %s\n",metric,x,y,w,h,style);
        if (!strcmp(metric,"pixel"))
          kr.metric=kate_pixel;
        else if (!strcmp(metric,"percent"))
          kr.metric=kate_percentage;
        else
          fprintf(stderr,"Invalid metric: %s\n",metric);
        kr.x=atoi(x);
        kr.y=atoi(y);
        kr.w=atoi(w);
        kr.h=atoi(h);
        kr.style=atoi(style);
        kate_info_add_region(&ki,&kr);
      }
      else if (starts(str,":region:")) {
        flush_headers();

        size_t region;
        sscanf(str,":region: %u\n",&region);
        kate_encode_set_region(&k,region);
      }
      else if (starts(str,":motion_test:")) {
        use_motion=1;
      }
      else if (starts(str,":text:")) {
        flush_headers();

        double t0,t1;
        int m0,s0,h0,m1,s1,h1;
        char text[4096];
        sscanf(str,":text: %02d:%02d.%02d %02d:%02d.%02d %[^\n]*\n",&m0,&s0,&h0,&m1,&s1,&h1,text);
        t0=m0*60+s0+h0/100.0;
        t1=m1*60+s1+h1/100.0;
        //fprintf(stderr,"timestamp: %d %d %d -> %f, %d %d %d -> %f\n",m0,s0,h0,t0,m1,s1,h1,t1);
        kate_encode_text(&k,t0,t1,use_motion?&test_motion:NULL,text,strlen(text),&op);
        ogg_stream_packetin(&os,&op);
        poll_page();
        use_motion=0;
      }
    }
    fgets(str,sizeof(str),stdin);
  }

  kate_encode_finish(&k,&op);
  ogg_stream_packetin(&os,&op);
  poll_page();

  flush_page();
  ogg_stream_clear(&os);
  kate_clear(&k);

  return 0;
}
